# Copyright 2021 Facundo Batista
# Licensed under the GPL v3 License
# For further info, check https://github.com/facundobatista/pyempaq

"""Main packer module."""

import argparse
import json
import pathlib
import shutil
import subprocess
import sys
import tempfile
import uuid
import venv
import zipapp
from collections import namedtuple

from pyempaq.config_manager import load_config, ConfigError

# collected arguments
Args = namedtuple("Args", "project_name basedir entrypoint requirement_files")


class ExecutionError(Exception):
    """The subprocess didn't finish ok."""


def find_venv_bin(basedir, exec_base):
    """Heuristics to find the pip executable in different platforms."""
    bin_dir = basedir / "bin"
    if bin_dir.exists():
        # linux-like environment
        return bin_dir / exec_base

    bin_dir = basedir / "Scripts"
    if bin_dir.exists():
        # windows environment
        return bin_dir / f"{exec_base}.exe"

    raise RuntimeError(f"Binary not found inside venv; subdirs: {list(basedir.iterdir())}")


def logged_exec(cmd):
    """Execute a command, redirecting the output to the log."""
    cmd = list(map(str, cmd))
    print(f"Executing external command: {cmd}")
    try:
        proc = subprocess.Popen(
            cmd, stdout=subprocess.PIPE, stderr=subprocess.STDOUT, universal_newlines=True)
    except Exception as err:
        raise ExecutionError(f"Command {cmd} crashed with {err!r}")
    stdout = []
    for line in proc.stdout:
        line = line[:-1]
        stdout.append(line)
        print(f":: {line}")
    retcode = proc.wait()
    if retcode:
        raise ExecutionError(f"Command {cmd} ended with retcode {retcode}")
    return stdout


def get_pip():
    """Ensure an usable version of `pip`."""
    useful_pip = pathlib.Path("pip3")
    # try to see if it's already installed
    try:
        logged_exec([useful_pip, "--version"])
    except ExecutionError:
        # failed to find a runnable pip, we need to install one
        pass
    else:
        return useful_pip

    # no useful pip found, let's create a virtualenv and use the one inside
    tmpdir = pathlib.Path(tempfile.mkdtemp())
    venv.create(tmpdir, with_pip=True)
    useful_pip = find_venv_bin(tmpdir, "pip3")
    return useful_pip


def pack(config):
    """Pack."""
    tmpdir = pathlib.Path(tempfile.mkdtemp())
    print(f"DEBUG packer: working in temp dir {str(tmpdir)!r}")

    # copy all the project content inside "orig" in temp dir
    origdir = tmpdir / "orig"
    shutil.copytree(config.basedir, origdir)

    # copy the unpacker as the entry point of the zip
    unpacker_src = pathlib.Path(__file__).parent / "unpacker.py"
    unpacker_final_main = tmpdir / "__main__.py"
    shutil.copy(unpacker_src, unpacker_final_main)

    # build a dir with the dependencies needed by the unpacker
    print("DEBUG packer: building internal dependencies dir")
    venv_dir = tmpdir / "venv"
    pip = get_pip()
    cmd = [pip, "install", "appdirs", f"--target={venv_dir}"]
    logged_exec(cmd)

    # store the needed metadata
    print("DEBUG packer: saving metadata from config", config)
    metadata = {
        "requirement_files": [str(path) for path in config.requirements],
        "project_name": config.name,
        "exec_default_args": config.exec.default_args,
    }
    if config.exec.script is not None:
        metadata["exec_style"] = "script"
        metadata["exec_value"] = str(config.exec.script)
    elif config.exec.module is not None:
        metadata["exec_style"] = "module"
        metadata["exec_value"] = str(config.exec.module)
    elif config.exec.entrypoint is not None:
        metadata["exec_style"] = "entrypoint"
        metadata["exec_value"] = str(config.exec.entrypoint)

    # if dependencies, store them just as another requirement file (save it inside the project,
    # but using an unique name to not overwrite anything)
    if config.dependencies:
        unique_name = f"pyempaq-autoreq-{uuid.uuid4()}.txt"
        extra_deps = origdir / unique_name
        extra_deps.write_text("\n".join(config.dependencies))
        metadata["requirement_files"].append(unique_name)

    metadata_file = tmpdir / "metadata.json"
    with metadata_file.open("wt", encoding="utf8") as fh:
        json.dump(metadata, fh)

    # create the zipfile
    packed_filepath = f"{config.name}.pyz"
    zipapp.create_archive(tmpdir, packed_filepath)

    # clean the temporary directory
    shutil.rmtree(tmpdir)

    print("Done, project packed in", packed_filepath)


def main():
    """Manage CLI interaction and call pack."""
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "source", type=pathlib.Path,
        help="The source file (pyempaq.yaml) or the directory where to find it.")
    args = parser.parse_args()
    try:
        print(f"Parsing configuration in {str(args.source)!r}")
        config = load_config(args.source)
    except ConfigError as err:
        print(err, file=sys.stderr)
        for err in err.errors:
            print(err, file=sys.stderr)
        exit(1)

    pack(config)



# Copyright 2021 Facundo Batista
# Licensed under the GPL v3 License
# For further info, check https://github.com/facundobatista/pyempaq

"""Init file to allow execution of PyEmpaq as a module."""

import sys

from pyempaq import main

sys.exit(main.main())



# Copyright 2021 Facundo Batista
# Licensed under the GPL v3 License
# For further info, check https://github.com/facundobatista/pyempaq

"""The configuration manager."""

import pathlib
from typing import List

import pydantic
import yaml


# base directory to which the different included paths may be relative from
_BASEDIR = None

# directory where the config is taken from, which is the default for basedir
_CONFIGDIR = None


class ConfigError(Exception):
    """Specific errors found in the config."""

    def __init__(self, msg=None, errors=None):
        if msg is None:
            msg = "Problem(s) found in configuration file:"
        super().__init__(msg)
        if errors is None:
            errors = []
        self.errors = errors


class ModelConfigDefaults(
        pydantic.BaseModel, extra=pydantic.Extra.forbid, frozen=True, validate_all=True):
    """Define defaults for the BaseModel configuration."""


class CustomStrictStr(pydantic.StrictStr):
    """Generic class to create custom strict strings validated by pydantic."""

    @classmethod
    def __get_validators__(cls):
        """Yield the relevant validators."""
        yield from super().__get_validators__()
        yield cls.custom_validate


class RelativePath(CustomStrictStr):
    """Constrained string which must be a relative path."""

    @classmethod
    def custom_validate(cls, value):
        """Validate."""
        if value is None:
            return
        value = pathlib.Path(value)

        if value.is_absolute():
            raise ValueError("path must be relative")

        # relative to the basedir
        abs_path = _BASEDIR / value

        if not abs_path.exists():
            raise ValueError(f"path {str(abs_path)!r} not found")

        if _BASEDIR not in abs_path.resolve().parents:
            raise ValueError("relative path must be inside the packed project")

        # return the relative path
        return value


class RelativeFile(RelativePath):
    """Constrained relative path which must be a file."""

    @classmethod
    def custom_validate(cls, value):
        """Validate."""
        value = super().custom_validate(value)
        if value is None:
            return

        # relative to the basedir
        abs_path = _BASEDIR / value

        if not abs_path.is_file():
            raise ValueError(f"path {str(abs_path)!r} must be a file")

        # return the relative path
        return value


class Executor(ModelConfigDefaults, alias_generator=lambda s: s.replace("_", "-")):
    """Executor information."""

    script: RelativeFile = None
    module: RelativePath = None
    entrypoint: List[pydantic.StrictStr] = None
    default_args: List[pydantic.StrictStr] = []


class Config(ModelConfigDefaults, validate_all=False):
    """Definition of PyEmpaq's configuration."""

    name: str
    basedir: pathlib.Path  # the default for this is managed on YAML load time
    exec: Executor
    requirements: List[RelativeFile] = []
    dependencies: List[str] = []

    @pydantic.validator("basedir")
    def ensure_basedir(cls, value):
        """Ensure that the basedir is valid, and store it to be used by other paths."""
        global _BASEDIR

        value = value.expanduser()
        if not value.is_absolute():
            value = _CONFIGDIR / value

        # set the basedir after the expanding/absolutizing, so the rest of the config can use it
        _BASEDIR = value

        if not value.exists():
            raise ValueError(f"path {str(value)!r} not found")
        if not value.is_dir():
            raise ValueError(f"path {str(value)!r} must be a directory")
        return value

    @pydantic.validator("exec", pre=True)
    def validate_exec_subkeys(cls, values):
        """Check the exec subkeys."""
        # it must be one, and only one, of these...
        subkeys = ["script", "module", "entrypoint"]
        count = sum(x in values for x in subkeys)
        if count == 0:
            subkeys_str = ', '.join(repr(x) for x in subkeys)
            raise ValueError(f"need at least one of these subkeys: {subkeys_str}")
        if count > 1:
            subkeys_str = ', '.join(repr(x) for x in subkeys)
            raise ValueError(f"only one of these subkeys is allowed: {subkeys_str}")
        return values


def _format_pydantic_errors(errors):
    """Format pydantic errors for a simpler presentation."""
    formated_errors = []
    for error in errors:
        # format location
        loc_parts = []
        for part in error['loc']:
            if isinstance(part, int):
                # an index, fix previous part
                loc_parts[-1] = f"{loc_parts[-1]}[{part}]"
            else:
                loc_parts.append(str(part))
        location = ".".join(loc_parts)

        message = error['msg'].strip()
        formated_errors.append(f"- {location!r}: {message}")

    return formated_errors


def load_config(path):
    """Load the config from charmcraft.yaml in the indicated directory."""
    global _CONFIGDIR

    path = path.expanduser().absolute()
    if path.is_dir():
        _CONFIGDIR = path
        configpath = path / "pyempaq.yaml"
    else:
        _CONFIGDIR = path.parent
        configpath = path

    if not configpath.exists():
        raise ConfigError(f"Configuration file not found: {str(configpath)!r}")

    try:
        content = yaml.safe_load(configpath.read_text())
    except Exception:
        raise ConfigError(f"Cannot open and parse YAML configuration file {str(configpath)!r}")

    # if `basedir` defaults to the directory where the configuration exists
    if "basedir" not in content:
        content["basedir"] = configpath.parent

    try:
        parsed = Config.parse_obj(content)
    except pydantic.error_wrappers.ValidationError as error:
        raise ConfigError(errors=_format_pydantic_errors(error.errors()))

    return parsed



# Copyright 2021 Facundo Batista
# Licensed under the GPL v3 License
# For further info, check https://github.com/facundobatista/pyempaq

"""Unpacking functionality.."""

import json
import os
import pathlib
import subprocess
import sys
import time
import venv
import zipfile

# this is the directory for the NEW virtualenv created for the project (not the packed
# one to run unpacker itself)
PROJECT_VENV_DIR = "project_venv"


def find_venv_bin(basedir, exec_base):
    """Heuristics to find the pip executable in different platforms."""
    bin_dir = basedir / "bin"
    if bin_dir.exists():
        # linux-like environment
        return bin_dir / exec_base

    bin_dir = basedir / "Scripts"
    if bin_dir.exists():
        # windows environment
        return bin_dir / "{}.exe".format(exec_base)

    raise RuntimeError("Binary not found inside venv; subdirs: {}".format(list(basedir.iterdir())))


def log(template, *args):
    """Print debug lines if proper envvar is present."""
    print("::pyempaq::", template.format(*args))


def get_python_exec(project_dir):
    """Return the Python exec to use.

    If a venv is present (just created or from a previous unpack) use it, else just use
    the one used to run this script.
    """
    venv_dir = project_dir / PROJECT_VENV_DIR
    if venv_dir.exists():
        executable = find_venv_bin(venv_dir, "python")
    else:
        executable = sys.executable
    return executable


log("Pyempaq start")

# parse pyempaq metadata from the zip file
pyempaq_filepath = pathlib.Path.cwd() / sys.argv[0]
zf = zipfile.ZipFile(pyempaq_filepath)
metadata = json.loads(zf.read("metadata.json").decode("utf8"))
log("Loaded metadata: {}", metadata)

# load appdirs from the builtin venv
sys.path.insert(0, f"{pyempaq_filepath}/venv/")
import appdirs  # NOQA: this is an import not at top of file because paths needed to be fixed

pyempaq_dir = pathlib.Path(appdirs.user_data_dir()) / 'pyempaq'
pyempaq_dir.mkdir(parents=True, exist_ok=True)
log("Temp base dir: {!r}", str(pyempaq_dir))

# create a temp dir and extract the project there
timestamp = time.strftime("%Y%m%d%H%M%S", time.gmtime(pyempaq_filepath.stat().st_ctime))
project_dir = pyempaq_dir / "{}-{}".format(metadata["project_name"], timestamp)
original_project_dir = project_dir / "orig"
if project_dir.exists():
    log("Reusing project dir {!r}", str(project_dir))
else:
    log("Creating project dir {!r}", str(project_dir))
    project_dir.mkdir()

    log("Extracting pyempaq content")
    zf.extractall(path=project_dir)

    venv_requirements = metadata["requirement_files"]
    if venv_requirements:
        log("Creating payload virtualenv")
        venv_dir = project_dir / PROJECT_VENV_DIR
        venv.create(venv_dir, with_pip=True)
        pip_exec = find_venv_bin(venv_dir, "pip3")
        cmd = [str(pip_exec), "install"]
        for req_file in venv_requirements:
            cmd += ["-r", str(original_project_dir / req_file)]
        log("Installing dependencies: {}", cmd)
        subprocess.run(cmd, check=True)
        log("Virtualenv setup finished")

    else:
        log("Skipping virtualenv (no requirements)")

python_exec = str(get_python_exec(project_dir))
os.chdir(original_project_dir)

if metadata["exec_style"] == "script":
    cmd = [python_exec, metadata["exec_value"]]
elif metadata["exec_style"] == "module":
    cmd = [python_exec, "-m", metadata["exec_value"]]
elif metadata["exec_style"] == "entrypoint":
    cmd = [python_exec] + metadata["exec_value"]
cmd.extend(metadata["exec_default_args"])

log("Running payload: {}", cmd)
subprocess.run(cmd)
log("Pyempaq done")



#!/usr/bin/env python3

# Copyright 2021 Facundo Batista
# Licensed under the GPL v3 License
# For further info, check https://github.com/facundobatista/pyempaq

"""Setup script for PyEmpaq."""

from setuptools import setup

with open("requirements.txt", "rt") as fh:
    install_requires = [x.strip() for x in fh]

setup(
    name="pyempaq",
    version="0.2.1",
    author="Facundo Batista",
    author_email="facundo@taniquetil.com.ar",
    description="A Python packer to run anywhwere any project with any virtualenv dependencies.",
    long_description=open("README.md", "rt", encoding="utf8").read(),
    long_description_content_type="text/markdown",
    url="https://github.com/facundobatista/pyempaq",
    license="GPL-v3",
    packages=["pyempaq"],
    classifiers=[
        "Environment :: Console",
        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
        "Operating System :: OS Independent",
        "Programming Language :: Python :: 3",
        "Topic :: System :: Archiving :: Packaging",
        "Topic :: System :: Software Distribution",
    ],
    entry_points={
        "console_scripts": ["pyempaq = pyempaq.main:main"],
    },
    install_requires=install_requires,
    python_requires=">=3.5",
)
